from abc import abstractmethod

from typing import List
from collections.abc import Iterable

class ResponseBuilder:
    @abstractmethod
    def build(self):
        pass

class MDHResponseBuilder(ResponseBuilder):
    def __init__(self,**kwargs):
        super().__init__()
        self.response = {}
        self._success = True
        for key,value in kwargs.items():
            print("{0} = {1}".format(key,value))
            self.response[key] = value

    def build(self):
        self.response['success'] = self._success
        return self.response

class DictResponseBuilder(MDHResponseBuilder):
     def __init__(self,d,**kwargs):
         super().__init__(**kwargs)
         self.data = d
         self.response['success'] = True
         self.response['data'] = self.data

     #def build(self):
     #    self.response['success'] = True
     #    self.response['data'] = self.data
     #    return self.response


class MDHErrorResponseBuilder(MDHResponseBuilder):
    def __init__(self,message='Something went wrong'):
        super().__init__()
        self._success = False
        self.message = message

    def build(self):
        self.response['success'] = self._success
        self.response['error'] = {
                'message': self.message
        }
        return self.response


class SQLAlchemyResponseBuilder(MDHResponseBuilder):
    def __init__(self,model_data=None,success=True):
        super().__init__()
        self.model_obj = model_data
        self._success = success
        self._only = None
        self._exclude = None
        self._camelKeys = True
        self._includeHybrid=True

#        self.set_model_data(model_data)

    def set_model_data(self, model_obj):
        if isinstance(model_obj, List):
            self.response['values'] = [el._to_dict() for el in model_obj]
        
        else:
            print("Model_obj: ",model_obj)
            data = model_obj._to_dict()
            
            if isinstance(data, List):
                self.response['values'] = data
            else:
                self.response['value'] = data

        return self

    def camel(self,camelKeys):
        self._camelKeys = camelKeys
        return self

    def only(self,_only):
        self._only = _only
        return self

    def exclude(self,_exclude):
        self._exclude = _exclude
        return self

    def build_model_object(self,model_obj):
            return model_obj._to_dict(
                    only=self._only,
                    exclude=self._exclude,
                    camelKeys=self._camelKeys,
                    includeHybrid=self._includeHybrid
            )
    def success(self):
        self._success = True
        self._error = False

    def error(self,msg):
        self._error = True
        self._success = False
        self._error_msg = msg

    def build_model_data(self):
        if self.model_obj:
            if isinstance(self.model_obj,Iterable):
                return [ self.build_model_object(o) for o in self.model_obj]
            return self.build_model_object(self.model_obj)
        else:
            return {}

    def build(self):
        self.response['data'] = self.build_model_data()
        self.response['success'] = self._success
        return self.response



from mdhwebserver.helpers.infixtopostfix import infix_to_postfix,postfix_calculate

class ExpertGraphRespsoneBuilder(MDHResponseBuilder):
    def __init__(self,model_data=None,success=True,expression=None):
        super().__init__()
        self.model_obj = model_data
        self._success = success

        self.expression = expression

    def build_response(self):
        
        dict_obj= self.model_obj._to_dict(
            only=['id','name',{'items':[{'parameters':['name','display_name']},'name','color','expression']}],
            camelKeys=True,
            includeHybrid=True
        )

        for idx,item in enumerate(self.model_obj.items):
            parameters = item.parameters
            ps_stack = infix_to_postfix(item.expression)

            points=[];
            for key,point in item.data_points_mapped.items():
                print("Mapping key: ",key)
                vallist = {}
                for param in parameters:
                    print("param is: ",param)
                    print("param name is: ",param.name)
                    val=getattr(point[param.name],"value")
                    print("Value is: ", val)
                    vallist[param.name] = val
                print("val list: ",vallist)
                res = postfix_calculate(ps_stack,vallist)
                print("Result = ",res)    
                points.append({'timestamp':key,'value':res})
            dict_obj["items"][idx]["dataPoints"] = points

        return dict_obj
        
    def build(self):
        self.response['data'] = self.build_response()
        self.response['success'] = self._success
        return self.response


