import logging
from datetime import datetime, timezone
from typing import Dict

import bcrypt
from sqlalchemy.dialects.postgresql import JSONB, JSON
from sqlalchemy import Column, String, Integer, Boolean, TIMESTAMP, Float, BigInteger, DDL, event, PrimaryKeyConstraint, Index, ARRAY, ColumnElement
from sqlalchemy.sql.schema import ForeignKey,UniqueConstraint
from sqlalchemy.orm import relationship,backref
from sqlalchemy import Table
from sqlalchemy import select
from sqlalchemy.orm import attribute_keyed_dict
from sqlalchemy.sql import operators
from sqlalchemy.orm import column_property, query_expression

#from ..database.db_config import db
#from mdhwebserver.appservices import database as db
from mdhwebserver.models.mdhbasemodel import BaseModel

from ..exceptions.pumpsense_error import MalformedPayloadError
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm.collections import (
    attribute_mapped_collection,
)

from .types.SiUnit import SiUnit
from sqlalchemy.sql import operators
from sqlalchemy.orm import column_property, query_expression
from .PumpsenseBase import Pumpsense

# class Pumpsense(BaseModel):
#     __abstract__ = True

#     id = Column(Integer, primary_key=True)
#     created_on = Column(TIMESTAMP, nullable=False, default=datetime.utcnow)


class User(Pumpsense):
    __tablename__ = 'users'
    _hidden = ['_password', 'password']

    logger = logging.getLogger('User')

    name = Column(String(128), nullable=False)
    email = Column(String(128), nullable=False, unique=True)
    company_name = Column(String(128), nullable=True, default=None)
    company_id = Column(String(128), nullable=True, default=None)
    _password = Column('password', String(256), nullable=False)
    active = Column(Boolean, nullable=False, default=True)
    authenticated = Column(Boolean, nullable=False, default=False)
    updated_on = Column(TIMESTAMP, nullable=False, default=datetime.utcnow, onupdate=datetime.utcnow)
    is_admin = Column(Boolean, nullable=False, default=False)
    unit_system = Column(String(4), server_default='fps')
    heatmap_view_id = Column(Integer,ForeignKey('heatmap_items.id',ondelete='SET NULL'))
    remote_token = Column(String(64),nullable=True)
    warn_enabled = Column(Boolean,server_default="t")

    heatmap_view = relationship("HeatmapItem", uselist=False)

    def __init__(self, name, email, password, company_name=None, company_id=None, is_admin=False) -> None:
        super().__init__()
        self.name = name
        self.email = email
        self.company_name = company_name
        self.company_id = company_id
        self.password = password
        self.is_admin = is_admin

    @hybrid_property
    def password(self):
        return self._password.encode('utf-8')
    
    @password.setter
    def password(self, value: bytes):
        self._password = bcrypt.hashpw(value, bcrypt.gensalt()).decode('utf-8')
    
    def check_password(self, entered_passwd: str):
        self.logger.debug('Inside check_password hybrid method')
        return bcrypt.checkpw(entered_passwd.encode('utf-8'), self.password)

    def is_active(self):
        return self.active

    def get_id(self):
        return self.email

    def is_authenticated(self):
        return self.authenticated

    def is_anonymous(self):
        return False

class ProductType(Pumpsense):
    __tablename__ = 'product_types'
    id = Column(Integer, primary_key=True)
    name = Column(String(128), nullable=False)
#    assemblies = relationship("Assembly",backref="product_type")
#    gauges = relationship("Gauge", backref='gauge_type')



assemblies_subassembly_instances_table = Table('assemblies_subassembly_instances', Pumpsense.metadata,
    Column('assembly_id',ForeignKey('assemblies.id', ondelete='CASCADE'),primary_key=True),
    Column('subassemblies_id',ForeignKey('subassembly_instances.id', ondelete='CASCADE'),primary_key=True)
)

class Assembly(Pumpsense):
    __tablename__ = 'assemblies'
    id = Column(Integer, primary_key=True)
#    product_type_id = Column(Integer, ForeignKey(ProductType.id),nullable=False)
    name = Column(String(128), nullable=False, unique=True)
    display_name = Column(String(128),nullable=True)
    image = Column(String(128), server_default="default_assembly_image")
    buildable = Column(Boolean, nullable=False, default=False)
    gauges = relationship("Gauge", backref='assembly')
    subassemblies = relationship("SubassemblyInstance", secondary=assemblies_subassembly_instances_table, back_populates='assemblies')
    subassemblies_dict = relationship("SubassemblyInstance", 
        secondary=assemblies_subassembly_instances_table, 
        collection_class=attribute_keyed_dict("name"),
        overlaps="subassemblies"
    )

    expert_data_groups = relationship("ExpertDataGroup", backref='assemblies')

 #   gauges = relationship("Gauge", backref='gauge_type')

class Datahub(Pumpsense):
    __tablename__ = 'datahubs'

    logger = logging.getLogger('Datahub')

    name = Column(String(128), nullable=True, unique=True)
    mac_address = Column(String(64), nullable=False, unique=True)
    ip_address = Column(String(64), nullable=True)
    onboarded = Column(Boolean, nullable=False, default=False)
    updated_on = Column(TIMESTAMP, nullable=False, default=datetime.utcnow, onupdate=datetime.utcnow)
 
    current_version = Column(String(20),nullable=True)
    assigned_version = Column(String(20),nullable=True)
    hub_id = Column(String(256),nullable=True)
    base_url = Column(String(256),nullable=True)
    status = Column(String(256),nullable=True)
    cloud_id = Column(String(50),nullable=True, unique=True)
    serial_no = Column(String(15), nullable=True, unique=True)
    machines = relationship("Gauge", backref='datahub')
    machines_mapped = relationship("Gauge", collection_class=attribute_mapped_collection('machine_id'),overlaps="datahub,machines")
    ota_downloaded_on = Column(TIMESTAMP, nullable=True)

    # Add the ota_file_id column and the foreign key relationship
    ota_file_id = Column(Integer, ForeignKey('ota_files.id'), nullable=True)
    
    # Define a relationship to the OtaFile class
    ota_file = relationship("OtaFile", backref='datahubs')

    
    #machines = relationship("Machine", backref='datahub')
    #machines_mapped = relationship("Machine", collection_class=attribute_mapped_collection('machine_id'))
    #new_field = Column(String(256),nullable=True)

#    def __repr__(self):
#        return '<%r %s>' % (self.__tablename__, self.name)


    @hybrid_property
    def last_checkin_duration(self):
        now = datetime.now(timezone.utc).timestamp()
        difference = (now - (self.updated_on.timestamp() + self.get_timezone_offset()) )
        return difference 
    
    @hybrid_property
    def since_last_update(self):
        now = datetime.now(timezone.utc).timestamp()
        difference = (now - (self.updated_on.timestamp() + self.get_timezone_offset()) )
        return difference 


class GaugeType(Pumpsense):
    __tablename__ = 'gauge_types'
    id = Column(Integer, primary_key=True)
    name = Column(String(128), nullable=False, unique=True)
    display_name = Column(String(128), nullable=True)
    image = Column(String(256),nullable=True)
    editable = Column(Boolean, nullable=False, server_default='t')
class SubassemblyInstance(Pumpsense):
    __tablename__ = 'subassembly_instances'

    name = Column(String(128), nullable=False, unique=True)
    display_name = Column(String(128), nullable=False)

    updated_on = Column(TIMESTAMP, nullable=False, default=datetime.utcnow, onupdate=datetime.utcnow)
#    pm_functions = relationship("PmFunction",backref=backref("subassembly",lazy="joined") )
    pm_functions = relationship("PmFunction",backref="subassembly")
    pm_functions_dict = relationship("PmFunction", collection_class=attribute_keyed_dict("name"), overlaps="pm_functions,subassembly")
    assemblies = relationship("Assembly", secondary=assemblies_subassembly_instances_table, back_populates='subassemblies', overlaps="subassemblies_dict")

#    def __repr__(self):
#        return '<%r %s>' % (self.__tablename__, self.name)
    
class GaugeGroup(Pumpsense):
    __tablename__ = 'gauge_groups'
    id = Column(Integer, primary_key=True)
    created_on = Column(TIMESTAMP, default=datetime.utcnow)
    name = Column(String(128), nullable=False)
    display_name = Column(String(128), nullable=False)
    pm_functions = relationship("PmFunction",back_populates='gauge_group')

from mdhwebserver.mixins.ListLike import ListLike, DictLike
from mdhwebserver.mixins.KeyedListCollection import KeyedListCollection,KeyedDictCollection
from .SiUnitModels import UnitTranslate
from sqlalchemy import SQLColumnExpression
import sqlalchemy as sa
import sqlalchemy.orm as orm

pm_functions_trend_meta_data_table = sa.Table('pm_functions_trend_meta_datas', Pumpsense.metadata,
    sa.Column('pm_function_id',sa.ForeignKey('pm_functions.id', ondelete='CASCADE'),primary_key=True),
    sa.Column('trend_meta_data_id',sa.ForeignKey('trend_meta_data.id', ondelete='CASCADE'),primary_key=True)
)



class PmFunction(Pumpsense):
    __tablename__ = 'pm_functions'

    id = Column(Integer, primary_key=True)
    name = Column(String(128), nullable=False)
    
    display_name = Column(String(128), nullable=False)
    subassembly_instance_id = Column(Integer, ForeignKey('subassembly_instances.id', ondelete="CASCADE"), index=True)
    gauge_group_id = Column(Integer, ForeignKey(GaugeGroup.id))
    updated_on = Column(TIMESTAMP, nullable=False, default=datetime.utcnow, onupdate=datetime.utcnow)
    trend_existence = Column(Boolean, nullable=True, default=False)
    mro = Column(Boolean, nullable=True, default=False)
    gauge_type = Column(String(20),nullable=True)
    gauge_base_image = Column(String(256),nullable=True)
    unit_system = Column(String(4),nullable=True,default="mks")
    units = Column(String(20))
    x_unit = Column(String(20), nullable=True)
    y_unit = Column(String(20), nullable=True)
    y_trend_min = Column(Float,nullable=True)
    y_trend_max = Column(Float,nullable=True)
    set_color = Column(Boolean,server_default='f', nullable=False)
    gauge_style = Column(String(50),server_default='standard')
    etl_pm_function_id = Column(Integer, ForeignKey('pm_functions.id',ondelete="CASCADE"), nullable=True)
    show_trend = Column(String,nullable=True)

    UniqueConstraint('name','subassembly_instance_id')
    graph_config = relationship("GraphConfig")
    x_unit_translation = relationship("UnitTranslate",primaryjoin="PmFunction.x_unit==foreign(UnitTranslate.name)", uselist=False)
    y_unit_translation = relationship("UnitTranslate",primaryjoin="PmFunction.y_unit==foreign(UnitTranslate.name)", uselist=False)

    trend_meta_mapped = relationship("TrendMetaData", collection_class=attribute_keyed_dict("name"))
    trend_meta = relationship("TrendMetaData", collection_class=ListLike,overlaps="trend_meta_mapped")
    trend_meta_data_direct = relationship("TrendMetaData", collection_class=ListLike, overlaps="trend_meta,trend_meta_mapped")
    trend_meta_data = orm.relationship("TrendMetaData", secondary=pm_functions_trend_meta_data_table)
#    trends = relationship("GaugeTrend", primaryjoin='and_(PmFunction.id == GaugeTrend.pm_func_id)')
    gauge_group = relationship("GaugeGroup",uselist=False, back_populates="pm_functions")
    device_data_items = relationship("DeviceDataItem", back_populates="pm_function")
    #device_data_items = relationship("DeviceDataItem", primaryjoin='foreign(DeviceDataItem.pm_function_id) == PmFunction.id', backref='pm_function')
    unit_translation = relationship("UnitTranslate", primaryjoin="PmFunction.units==foreign(UnitTranslate.name)", uselist=False)
#    def __repr__(self):
#        return '<%r %s>' % (self.__tablename__, self.name)
    gauge_ranges = relationship('GaugeRange',back_populates='pm_function')

    etl_gauges = relationship('PmFunction',primaryjoin="PmFunction.id == PmFunction.etl_pm_function_id",back_populates='etl_gauge')
    etl_gauge = relationship('PmFunction',primaryjoin="PmFunction.etl_pm_function_id == remote(PmFunction.id)", uselist=False, back_populates='etl_gauges')
    mro_issues = relationship('MroIssue', back_populates="pm_function")

    @hybrid_property
    def subassembly_instance_name(self):
        return self.subassembly.name

    @subassembly_instance_name.inplace.expression
    @classmethod
    def _subassembly_instance_name_expression(cls) -> SQLColumnExpression[String]:  
    #    print("------------ subassembly_instance_name expression")  
    #    return select([SubassemblyInstance.name]).where(cls.subassembly_instance_id == SubassemblyInstance.id).scalar_subquery()
       return (select(SubassemblyInstance.name).where(cls.subassembly_instance_id == SubassemblyInstance.id).label("Xsubassembly_instance_name"))


class TrendMetaData(Pumpsense):
    __tablename__ = 'trend_meta_data'

    name = Column(String(128), nullable=False)
    pm_func_id = Column(Integer, ForeignKey('pm_functions.id',ondelete='CASCADE'), index=True)
    display_name = Column(String(128), nullable=False)
    color = Column(String(16), nullable=False)
    updated_on = Column(TIMESTAMP, nullable=False, default=datetime.utcnow, onupdate=datetime.utcnow)
    unit_data_points = relationship("DeviceUnitDataPoint", order_by="DeviceUnitDataPoint._timestamp", cascade="all,delete-orphan", back_populates="trend_meta_data")
    units = Column(String(20), nullable=True)          # Data incomming units
    fps_units = Column(String(20), nullable=True)      # FPS display units
    mks_units = Column(String(20), nullable=True)      # MKS display units
    gauge_ranges = relationship('GaugeRange',back_populates='trend_meta_data')
    pm_functions = orm.relationship("PmFunction", 
        secondary=pm_functions_trend_meta_data_table,
        # foreign_keys="pm_functions_trend_meta_data_table.c.pm_function_id"
    )
    pm_function = orm.relationship(PmFunction,uselist=False)

    device_data_items = relationship('DeviceDataItem',
      secondary="device_data_mappers"
        
    )


    # virtual_pm_functions = orm.relationship("PmFunction", 
    #     secondary=pm_functions_trend_meta_data_table,
    #     foreign_keys="pm_functions_trend_meta_data_table.c.pm_function_id",
    #     primaryjoin='and_(PmFunction.guage_type == "virtual")'
    # )
    # virtual_pm_functions = pm_functions.and_(PmFunction.guage_type == "virtual")

#    def __repr__(self):
#        return '<%r %s>' % (self.__tablename__, self.toJson())

gauges_device_data_items_table = Table('gauges_device_data_items', Pumpsense.metadata,
    Column('gauge_id',ForeignKey('gauges.id',ondelete='CASCADE'),primary_key=True),
    Column('device_data_item_id',ForeignKey('device_data_items.id',ondelete='CASCADE'),primary_key=True)
)

# Need to revisit what we want to happen on a parent/child being deleted.

gauges_child_parent_items = Table('gauges_child_parent_items', Pumpsense.metadata,
    Column('parent_id',ForeignKey('gauges.id',ondelete='CASCADE'),primary_key=True),
    Column('child_id',ForeignKey('gauges.id',ondelete='CASCADE'),primary_key=True)
)

from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.associationproxy import AssociationProxy
from typing import List

class Gauge(Pumpsense):
    __tablename__ = 'gauges'
#   __with__ = {"Gauge_configs"}

#   __mapper_args__ = {
#           'include_properties': ['id','machine_id' ]
#   }
    logger = logging.getLogger('Gauge')

    id = Column(Integer, primary_key=True)
    assembly_id = Column(Integer, ForeignKey(Assembly.id, ondelete="CASCADE"),nullable=False)

    pump_number = Column(Integer, nullable=True, default=None)
    machine_id = Column(String(64), nullable=False, unique=True)
    machine_name = Column(String(128), nullable=False)
    company_name = Column(String(128), nullable=True)
    location_name = Column(String(128), nullable=True)
 #  data = Column(JSONB, nullable=False)
    updated_on = Column(TIMESTAMP, nullable=False, default=datetime.now, onupdate=datetime.now)
    datahub_id = Column(Integer, ForeignKey(Datahub.id,ondelete="SET NULL"),nullable=True)
    onboarded = Column(Boolean, nullable=False, default=False)
    package_serial_no = Column(String(15), nullable=True, unique=True)
    color = Column(String(30),nullable=True)
    machine_view_type = Column(String(30), nullable=True, server_default='machine')

    child_ids: AssociationProxy[List[int]] = association_proxy("children", "id", creator=lambda children: Gauge(id=children))

#   subassemblies = relationship("SubassemblyInstance")

#   trends = relationship("GaugeTrend",primaryjoin='and_(Gauge.machine_id == GaugeTrend.machine_id)', backref='gauges')
#   trends = relationship("GaugeTrend",primaryjoin='and_(Gauge.machine_id == GaugeTrend.machine_id)', backref='gauge')

#    device_data_items = relationship("DeviceDataItem", foreign_keys="DeviceDataItem.gauge_id", backref='gauge')
    device_data_items = relationship("DeviceDataItem", secondary=gauges_device_data_items_table, back_populates='machines')

    device_unit_data_points = relationship("DeviceUnitDataPoint", foreign_keys='DeviceUnitDataPoint.gauge_id', backref='gauge')

    mro_logs = relationship("MroLog", backref='gauge')
   #machine = relationship("Machine", uselist=False, backref='gauge', foreign_keys=[machine_id], primaryjoin='Gauge.machine_id == Machine.machine_id')    # uselist makes it one to one
    gauge_groups = relationship("GaugeGroup", secondary="device_data_items", primaryjoin="Gauge.id==DeviceDataItem.gauge_id", secondaryjoin="GaugeGroup.id == DeviceDataItem.gauge_group_id",viewonly=True)
    heatmap_items = relationship("HeatmapItem", back_populates="gauge", cascade="all,delete-orphan")
    heatmap_item = relationship("HeatmapItem", primaryjoin="and_(foreign(HeatmapItem.name) == Gauge.machine_name)", back_populates="gauge",uselist=False)


    children = relationship("Gauge",
        secondary=gauges_child_parent_items,
        primaryjoin=gauges_child_parent_items.c.parent_id == id,
        secondaryjoin=id == gauges_child_parent_items.c.child_id,
        join_depth=2,
        back_populates="parents",
        overlaps="children,parents"
    )

    # children_dict = relationship("Gauge",
    #     secondary=gauges_child_parent_items,
    #     primaryjoin=gauges_child_parent_items.c.parent_id == id,
    #     secondaryjoin=id == gauges_child_parent_items.c.child_id,
    #     join_depth=2,
    #     collection_class=attribute_keyed_dict("name"),
    #     back_populates="parents_dict"
    # )

    parents = relationship("Gauge",
        secondary=gauges_child_parent_items,
        primaryjoin=gauges_child_parent_items.c.child_id == id,
        secondaryjoin=id == gauges_child_parent_items.c.parent_id,
        join_depth=2,
        back_populates="children",
        overlaps="children,parents"
        
    )
    # parents_dict = relationship("Gauge",
    #     secondary=gauges_child_parent_items,
    #     primaryjoin=gauges_child_parent_items.c.child_id == id,
    #     secondaryjoin=id == gauges_child_parent_items.c.parent_id,
    #     join_depth=2,
    #     back_populates="children_dict"
    # )

    # child_ids: AssociationProxy[List[int]] = association_proxy("children", "id", creator=lambda children: Gauge().where(id=children).first())
    # child_ids: AssociationProxy[List[int]] = association_proxy("children", "id")
    parent_association_recs = relationship("GaugeAssociation", back_populates="parent", foreign_keys="gauges_child_parent_items.c.parent_id", overlaps="children,parents")
    child_ids: AssociationProxy[List[int]] = association_proxy(
        "parent_association_recs", 
        "child_id", 
        creator=lambda children: GaugeAssociation(child_id=children)
    )

    def color_update(self):
        color_list = []
        color_list  = make_color_list(self.children) + make_color_list(self.device_data_items, lambda x: x.pm_function.mro == True)
        max_color = get_max_color(color_list)
        if self.color != max_color:
            self.color = max_color

    @hybrid_property
    def since_last_update(self):
        now = datetime.now(timezone.utc).timestamp()
        difference = (now - (self.updated_on.timestamp() + self.get_timezone_offset()) )
        return difference 


#   def __repr__(self):
#        return '<%r %s>' % (self.__tablename__, self.name)


#   @hybrid_property
#   def subassemblies(self):
#        return self.assembly.subassemblies

#   @hybrid_property
    def device_data_item(self,pm_func_id,subassembly_id):
        #print(f'Getting device data item for {pm_func_id} -- {subassembly_id}')
        #print(self.device_data_items)
        #print("Type is: ",type(self.device_data_items))
        for datum in self.device_data_items:
            #print("Datum id is: ",datum.id)
            if datum.pm_func_id == pm_func_id and datum.subassembly_instance_id == subassembly_id:
                return datum
#        if self.device_data:
#            return self.device_data[0]
        return None

#   @hybrid_property
#   def pm_func(self):
#        if self.accounts:
#            return self.Gauge_configs[0]
#        else:
#            return None

   #@hybrid_property
   #def onboarded(self):
   #    return self.machine.onboarded

   #@onboarded.expression
   #def onboarded(cls):
   #    #return select([Machine.onboarded]).where(cls.machine_id == Machine.machine_id).as_scaler()
   #    return select([Machine.machine_id]).where(cls.machine_id == Machine.machine_id)
    
class GaugeAssociation(BaseModel):
    __table__ = gauges_child_parent_items
    child = relationship("Gauge", backref="child_association_recs",
        foreign_keys="gauges_child_parent_items.c.child_id",
        overlaps="children,parents"
    )

    parent = relationship("Gauge", back_populates="parent_association_recs",
        foreign_keys="gauges_child_parent_items.c.parent_id",
        overlaps="children,parents"
    )

#class GaugeFunctionLink(Pumpsense):
#    id = Column(Integer, primary_key=True)
#    gauge_id = Column(db.Integer, db.ForeignKey('gauges.id',ondelete='CASCADE'))
#    pm_function_id = Column(db.Integer,db.ForeignKey('pm_functions.id'))

# device_data_item_data_mapper_table = sa.Table('device_data_mappers', Pumpsense.metadata,
#     sa.Column('device_data_item_id',sa.ForeignKey('device_data_items.id', ondelete='CASCADE'),primary_key=True),
#     sa.Column('trend_meta_data_id',sa.ForeignKey('trend_meta_data.id', ondelete='CASCADE'),primary_key=True),
#     sa.Column('gauge_id',sa.ForeignKey('gauges.id', ondelete='CASCADE'),primary_key=True)
# )

class DeviceDataItemUnitDataPointAssociation(Pumpsense):
    __tablename__ = 'device_data_mappers'

    device_data_item_id = Column(Integer, ForeignKey("device_data_items.id",ondelete='CASCADE'),nullable=False, index=True)
    trend_meta_data_id = Column(Integer, ForeignKey(TrendMetaData.id,ondelete='CASCADE'),nullable=False, index=True)
    gauge_id = Column(Integer, ForeignKey(Gauge.id,ondelete='CASCADE'),nullable=False, index=True)

    # device_data_item = relationship("DeviceDataItem", foreign_keys=[device_data_item_id])
    trend_meta_data = relationship(TrendMetaData)
    gauge = relationship(Gauge)
    device_unit_data_point=relationship("DeviceUnitDataPoint",
        primaryjoin="foreign(DeviceUnitDataPoint.trend_meta_data_id) == DeviceDataItemUnitDataPointAssociation.trend_meta_data_id and foreign(DeviceUnitDataPoint.gauge_id) == DeviceDataItemUnitDataPointAssociation.gauge_id",
        back_populates="device_data_item_associations"
    )
    device_data_item = relationship("DeviceDataItem",back_populates="device_unit_data_point_associations")

class DeviceDataItem(Pumpsense):
    __tablename__ = 'device_data_items'

    id = Column(Integer, primary_key=True)
    gauge_id = Column(Integer, ForeignKey(Gauge.id,ondelete='CASCADE'),nullable=False, index=True)
    # data = Column(JSONB, nullable=False)
#    pm_func_id = Column(Integer, ForeignKey('pm_functions.id'))
    pm_func_id = Column(Integer, ForeignKey(PmFunction.id, ondelete='CASCADE'), nullable=False, index=True)
    pm_function_name = Column(String(128), nullable=True)           # This should not be needed
    updated_on = Column(TIMESTAMP, nullable=False, default=datetime.utcnow, onupdate=datetime.utcnow)
#    subassembly_instance_id = Column(Integer, ForeignKey('subassembly_instances.id'))
    subassembly_instance_id = Column(Integer)
    subassembly_instance_name = Column(String(128), nullable=True)  # This should not be needed
    gauge_group_id = Column(Integer, ForeignKey(GaugeGroup.id), nullable=True)
    gauge_group_name = Column(String(128), nullable=True)
    colors = Column(ARRAY(String(20), as_tuple=False, dimensions=None, zero_indexes=False))
    values = Column(ARRAY(Float, as_tuple=False, dimensions=None, zero_indexes=False))
    reason = Column(String(512),nullable=True)
    activate_hourglass = Column(Boolean,nullable=True, default=False)
    feedback = Column(Boolean,nullable=True,default=False)
    user_input = Column(Boolean,nullable=True, default=False)
    hide_from_group = Column(Boolean,nullable=True, default=False)
    message = Column(String(256))
    yellow_threshold = Column(Float,nullable=True)
    red_threshold = Column(Float,nullable=True)
    UniqueConstraint(gauge_id,subassembly_instance_id,pm_func_id, name="device_data_item_uniq_idx")
    kpis = Column(JSON, nullable=True)
    color = Column(String(30),nullable=True)
    #mart = Column(JSONB, nullable=False)  # Need to figure this out, it is going away with KPIs.
    machines = relationship("Gauge", secondary=gauges_device_data_items_table, back_populates='device_data_items')
    pm_function = relationship(PmFunction, back_populates="device_data_items")
    heatmap_item = relationship("HeatmapItem",back_populates="device_data_item",uselist=False)
    gauge_ranges = relationship('GaugeRange',back_populates='device_data_item')
    mro_issues = relationship('MroIssue', primaryjoin="foreign(MroIssue.pm_function_id) == DeviceDataItem.pm_func_id")
    mro_color_issues = relationship('MroIssue', primaryjoin="and_(foreign(MroIssue.color) == DeviceDataItem.color,foreign(MroIssue.pm_function_id) == DeviceDataItem.pm_func_id)")
    trend_meta_data_direct = relationship(TrendMetaData, primaryjoin="DeviceDataItem.pm_func_id == remote(TrendMetaData.pm_func_id)", foreign_keys=pm_func_id, uselist=True)
    # trend_meta_data = orm.relationship("TrendMetaData", primaryjoin=pm_func_id == orm.remote(pm_functions_trend_meta_data_table.c.pm_function_id),secondary=pm_functions_trend_meta_data_table)
    # device_unit_data = orm.relationship("DeviceUnitDataPoint",secondary=device_data_item_data_mapper_table, secondaryjoin="and_(foreign(device_data_mappers.c.gauge_id)==DeviceUnitDataPoint.gauge_id)")
    device_unit_data_point_associations = relationship(DeviceDataItemUnitDataPointAssociation,back_populates="device_data_item")
    trend_meta_datas = relationship(TrendMetaData,secondary="device_data_mappers")
    
#class MyTimestamp(TIMESTAMP):
#    pass

class GraphConfig(Pumpsense):
    __tablename__ = 'graph_configs'
    id = Column(Integer, primary_key=True)
    pm_function_id = Column(Integer, ForeignKey('pm_functions.id'))
    unit_system = Column(String(4),nullable=True,default="mks")
    y_max = Column(Float, nullable=True)
    y_min = Column(Float, nullable=True)
    y_unit = Column(String,nullable=True)
    x_unit = Column(String,nullable=True)
    yellow_threshold = Column(Float,nullable=True)
    yellow_threshold_text = Column(String)
    red_threshold = Column(Float,nullable=True)
    red_threshold_text = Column(String)


class DeviceUnitDataPoint(Pumpsense):
    __tablename__ = 'device_unit_data_points'
    id = Column(Integer, autoincrement=True)
    #id = Column(Integer, primary_key=True)
    gauge_id = Column(Integer, ForeignKey(Gauge.id,ondelete='CASCADE'),nullable=False)
    # machine_id = Column(String(64), nullable=False)
    _timestamp = Column("timestamp",TIMESTAMP,nullable=False)
    converted = query_expression()
    mks = query_expression()
    fps = query_expression()
    #_timestamp = Column("timestamp",TIMESTAMP,nullable=False,primary_key=True)
#    _timestamp = Column(Integer)
    trend_meta_data_id = Column(Integer, ForeignKey(TrendMetaData.id,ondelete='CASCADE'),nullable=False,index=True)
    trend_meta_data = relationship(TrendMetaData, back_populates="unit_data_points", uselist=False)
    device_data_item_associations = relationship("DeviceDataItemUnitDataPointAssociation",
        primaryjoin="DeviceUnitDataPoint.trend_meta_data_id == remote(DeviceDataItemUnitDataPointAssociation.trend_meta_data_id) and DeviceUnitDataPoint.gauge_id == remote(DeviceDataItemUnitDataPointAssociation.gauge_id)",
        back_populates="device_unit_data_point",foreign_keys=[trend_meta_data_id,gauge_id]
    )

    value = Column(SiUnit)
    __table_args__ = (
        PrimaryKeyConstraint('id','timestamp'),
        {},
    )

    @hybrid_property
#    @property
    def timestamp(self):
        return self._timestamp.timestamp()

    @timestamp.setter
    def timestamp(self,value):
        self._timestamp = value
        #self._timestamp = datetime.datetime.utcfromtimestamp(value)

    @timestamp.expression
    def timestamp(cls):
        return cls._timestamp
#        return cls._timestamp

#    @timestamp.getter
#    def timestamp(self):
#        return self.timestamp
    

Index('unit_data_point_graphing_idx',DeviceUnitDataPoint.gauge_id,DeviceUnitDataPoint.timestamp,DeviceUnitDataPoint.trend_meta_data_id)
    
# event.listens(
#     DeviceUnitDataPoint.__table__,
#     'after_create',
#     DDL(f"SELECT create_hypertable('{DeviceUnitDataPoint.__tablename__}', 'timestamp', chunk_time_interval := INTERVAL '24 HOURS')")
# )

from sqlalchemy import text

@event.listens_for(DeviceUnitDataPoint.__table__,'after_create')
def device_data_after_create(target,connection,**kw):
    connection.execute(text(f"SELECT create_hypertable('{DeviceUnitDataPoint.__tablename__}', 'timestamp', chunk_time_interval := INTERVAL '24 HOURS')"))
    connection.execute(text(f"SELECT add_retention_policy('{DeviceUnitDataPoint.__tablename__}', INTERVAL '90 DAYS')"))



class GaugeTrend(Pumpsense):
    __tablename__ = 'gauge_trends'
    __abstract__ = True

    logger = logging.getLogger('GaugeTrend')

    machine_id = Column(String(64), ForeignKey('gauges.machine_id'),nullable=False)
#    machine_id = Column(String(64), nullable=False)
    pm_func_id = Column(Integer, ForeignKey('pm_functions.id'))
    data = Column(JSONB, nullable=False)


#    def __repr__(self):
#        return '<%r %s>' % (self.__tablename__, self.toJson())


#    @hybrid_property
#    def update_range(self,start,end):
#        return self.updated_on 
        
#    @update_range.expression
#    def update_range(cls,start,end):
#        print("Start is: ",start)
#        print("End is: ",end)
#        return (cls.created_on>= start and cls.created_on<= end)
'''
class Machine(Pumpsense):
    __tablename__ = 'machines'
    
    id = Column(Integer, primary_key=True)
    datahub_id = Column(Integer, ForeignKey('datahubs.id'))
    machine_id = Column(String(64), nullable=False)
    onboarded = Column(Boolean, nullable=False, default=False)
'''


                
# Expert data section
class ExpertDataGroup(Pumpsense):
    __tablename__ = 'expert_data_groups'

    id = Column(Integer, primary_key=True)
    name = Column(String(128), nullable=False)
    display_name = Column(String(128), nullable=False)
    assembly_id = Column(Integer, ForeignKey('assemblies.id'))
    expert_data_items = relationship("ExpertDataItem", backref='expert_data_group')
    mapped_expert_data_items = relationship("ExpertDataItem", 
        back_populates='expert_data_group2', 
        collection_class=attribute_mapped_collection('name'),
        overlaps='expert_data_group,expert_data_items'
    )

class ExpertDataItem(Pumpsense):
    __tablename__ = 'expert_data_items'

    id = Column(Integer, primary_key=True)
    expert_data_group_id = Column(Integer, ForeignKey('expert_data_groups.id'))
    name = Column(String(128), nullable=False)
    display_name = Column(String(128), nullable=False)
    expert_data_points = relationship("ExpertDataPoint", back_populates='expert_data_item')
    expert_data_group2 = relationship("ExpertDataGroup", back_populates="mapped_expert_data_items", overlaps="expert_data_group,expert_data_items")

class ExpertDataPoint(Pumpsense):
    __tablename__ = 'expert_data_points'
    
    #id = Column(Integer, primary_key=True)
    id = Column(Integer, autoincrement=True)
    expert_data_item_id = Column(Integer, ForeignKey('expert_data_items.id'))
    gauge_id = Column(Integer, ForeignKey(Gauge.id,ondelete='CASCADE'),nullable=False)
    machine_id = Column(String(64), nullable=False)
    _timestamp = Column("timestamp",TIMESTAMP)
    raw_timestamp = Column(BigInteger)
    __table_args__ = (
        PrimaryKeyConstraint('id','timestamp'),
        {},
    )

    value = Column(Float)
    
    expert_data_item = relationship(ExpertDataItem, back_populates='expert_data_points')

    @hybrid_property
    def timestamp(self):
        return round(self._timestamp.timestamp())

    @timestamp.setter
    def timestamp(self,value):
        self._timestamp = value
        #self._timestamp = datetime.datetime.utcfromtimestamp(value)

    @timestamp.expression
    def timestamp(cls):
        return cls._timestamp
#        return cls._timestamp

    @hybrid_property
    def data_point_name(self):
        return self.expert_data_item.name

Index('expert_data_point_graphing_idx',ExpertDataPoint.machine_id,ExpertDataPoint.timestamp,ExpertDataPoint.expert_data_item_id)
    
event.listen(
    ExpertDataPoint.__table__,
    'after_create',
    DDL(f"SELECT create_hypertable('{ExpertDataPoint.__tablename__}', 'timestamp')")
)
    
class ExpertReport(Pumpsense):
    __tablename__ = 'expert_reports'
    
    id = Column(Integer, primary_key=True)
    name = Column(String(128), nullable=False)
    display_name = Column(String(128), nullable=False)

    items = relationship("ExpertReportItem", backref='expert_report')


    
expert_report_items_expert_data_items = Table('expert_report_items_expert_data_items', Pumpsense.metadata,
    Column('expert_data_item_id',ForeignKey('expert_data_items.id')),
    Column('expert_report_item_id',ForeignKey('expert_report_items.id'))
)

class ExpertReportItem(Pumpsense):
    __tablename__ = 'expert_report_items'
    
    id = Column(Integer, primary_key=True)
    expert_report_id = Column(Integer, ForeignKey('expert_reports.id'))

    name = Column(String(128), nullable=False)
    color = Column(String(30),nullable=False)
    expression = Column(String(256),nullable=True)
    parameters = relationship("ExpertDataItem", secondary=expert_report_items_expert_data_items, backref='ExpertReportItems')
    data_points = relationship(ExpertDataPoint,order_by="ExpertDataPoint._timestamp", 
        viewonly=True,
        secondary=expert_report_items_expert_data_items,
        primaryjoin='and_(expert_report_items_expert_data_items.c.expert_report_item_id == ExpertReportItem.id)',
        secondaryjoin='foreign(ExpertDataPoint.expert_data_item_id) == expert_report_items_expert_data_items.c.expert_data_item_id'
    )
    data_points_mapped = relationship(ExpertDataPoint,order_by="ExpertDataPoint._timestamp", 
        viewonly=True,
        secondary=expert_report_items_expert_data_items,
        primaryjoin='and_(expert_report_items_expert_data_items.c.expert_report_item_id == ExpertReportItem.id)',
        secondaryjoin='foreign(ExpertDataPoint.expert_data_item_id) == expert_report_items_expert_data_items.c.expert_data_item_id',
        #collection_class=DictLike('timestamp')
        #collection_class=lambda: KeyedDictCollection('timestamp','data_point_name')
        #collection_class=lambda: KeyedListCollection('timestamp')
        #collection_class=lambda: DictLike('timestamp','expert_data_item_id')

        collection_class=lambda: DictLike('timestamp','data_point_name')
        #collection_class=DictLike('timestamp','expert_data_item_id')

    )


class SmtpConfig(Pumpsense):
    __tablename__ = 'smtp_config'
    
    id = Column(Integer, primary_key=True)
    smtpServer = Column(String(256),nullable=False)
    smtpPort = Column(Integer,nullable=False)
    smtpUsername = Column(String(256),nullable=True)
    smtpPassword = Column(String(256),nullable=True)
    fromName = Column(String(256),nullable=False)
    tls = Column(Boolean,nullable=False, default=True)
    fromEmail = Column(String(256),nullable=False)
    authenticate = Column(Boolean,nullable=False, default=True)

   
class MroLog(Pumpsense):
    __tablename__ = 'mro_logs'
    id = Column(Integer, primary_key=True)
    gauge_id = Column(Integer, ForeignKey(Gauge.id,ondelete='CASCADE'),nullable=False)
    machine_id = Column(String(64), nullable=False, index=True)
    _timestamp = Column("timestamp",TIMESTAMP,nullable=False)
    pm_function_id = Column(Integer, ForeignKey('pm_functions.id'))
    status = Column(String(30),nullable=True)
    color = Column(String(30),nullable=False)
    pm_function = relationship("PmFunction",backref="mro_log",uselist=False)

    @hybrid_property
    def timestamp(self):
        if self._timestamp:
            return round(self._timestamp.timestamp())
        else:
            return None

    @timestamp.setter
    def timestamp(self,value):
        self._timestamp = value
        #self._timestamp = datetime.datetime.utcfromtimestamp(value)


class MailLog(Pumpsense): 
    __tablename__ = 'mail_logs'
    id = Column(Integer, primary_key=True)
    sender = Column(String(255),nullable=False)
    to = Column(String(255),nullable=False)
    subject = Column(String(255),nullable=False)
    server = Column(String(255),nullable=False)
    port = Column(Integer,nullable=False)
    result = Column(Boolean,nullable=False)
    error = Column(String(255),nullable=True)


heatmap_child_parent_items = Table('heatmap_child_parent_items', Pumpsense.metadata,
    Column('parent_id',ForeignKey('heatmap_items.id',ondelete='CASCADE'),primary_key=True),
    Column('child_id',ForeignKey('heatmap_items.id',ondelete='CASCADE'),primary_key=True)
)

from mdhwebserver.utils.heatmap import get_max_color, make_color_list

class HeatmapItem(Pumpsense):
    __tablename__ = 'heatmap_items'
    id = Column(Integer, primary_key=True)
    # parent_id = Column(Integer, ForeignKey('heatmap_items.id'), nullable=True)
    gauge_id = Column(Integer, ForeignKey(Gauge.id,ondelete='CASCADE'), nullable=True, index=True)
    pm_function_id = Column(Integer, ForeignKey('pm_functions.id', ondelete='CASCADE'), nullable=True, index=True)
    device_data_item_id = Column(Integer, ForeignKey(DeviceDataItem.id,ondelete='CASCADE'), nullable=True, index=True)
    # group_id = Column(Integer, nullable=True)
    name = Column(String(128), nullable=False)
    color = Column(String(50), nullable=True)
    of_interest = Column(Boolean,nullable=False,default=True)
    updated_on = Column(TIMESTAMP, nullable=False, default=datetime.utcnow, onupdate=datetime.utcnow)
    pm_function = relationship("PmFunction", foreign_keys=[pm_function_id])
    type = Column(String(128), nullable=True)
    order_id = Column(Integer, nullable=True)
    color = Column(String(50), nullable=True)
    color_count = Column(Integer, nullable=False, default=0, server_default="0")
    alarm_limit = Column(Integer, nullable=False, default=1, server_default="1")
    last_color_update = Column(Integer,nullable=True)
    view_type = Column(String(30), nullable=True, index=True)

    gauge = relationship("Gauge", foreign_keys=[gauge_id], back_populates="heatmap_items")
    device_data_item = relationship("DeviceDataItem",uselist=False, back_populates="heatmap_item")

    rchildren = relationship("HeatmapItem",
        secondary=heatmap_child_parent_items,
        primaryjoin=heatmap_child_parent_items.c.parent_id == id,
        secondaryjoin=id == heatmap_child_parent_items.c.child_id,
        join_depth=2,
        back_populates="rparents",
        overlaps="rchildren,rparents"
    )

    # rchildren_dict = relationship("HeatmapItem",
    #     secondary=heatmap_child_parent_items,
    #     primaryjoin=heatmap_child_parent_items.c.parent_id == id,
    #     secondaryjoin=id == heatmap_child_parent_items.c.child_id,
    #     join_depth=2,
    #     collection_class=attribute_keyed_dict("name"),
    #     back_populates="rparents_dict",
    #     overlaps="rchildren,rparents",
        
    # )

    rparents = relationship("HeatmapItem",
        secondary=heatmap_child_parent_items,
        primaryjoin=heatmap_child_parent_items.c.child_id == id,
        secondaryjoin=id == heatmap_child_parent_items.c.parent_id,
        join_depth=2,
        back_populates="rchildren",
        overlaps="rchildren,rparents"
    )
    # rparents_dict = relationship("HeatmapItem",
    #     secondary=heatmap_child_parent_items,
    #     primaryjoin=heatmap_child_parent_items.c.child_id == id,
    #     secondaryjoin=id == heatmap_child_parent_items.c.parent_id,
    #     join_depth=2,
    #     back_populates="rchildren_dict",
    #     overlaps="rchildren,rparents"
    # )

    def sync_machine(self,machine):
        print("------------------------ Syncing machine")
        self.rchildren=[]
        for child in machine.children:
            print("---------- Have child: ",child)
            if not child.heatmap_item:
                print("Child not in heatmap")
                continue
            self.rchildren.append(child.heatmap_item)
        for device_data_item in machine.device_data_items:
            print("---------- Have Device data item: ",device_data_item)
            if not device_data_item.heatmap_item:
                print("Child not in heatmap")
                continue
            self.rchildren.append(device_data_item.heatmap_item)
        self.color_update()

    def color_update(self, propagate=False, ts=None):
        color_list = make_color_list(self.rchildren)
        max_color = get_max_color(color_list)
        if max_color != self.color:
            self.color = max_color
            self.color_count=0
            if ts is not None:
                self.last_color_update = ts
        else:
            self.color_count += 1
        
        if propagate:    # If we are going to pass new color up to parent.  Just stubbed for now.
            if self.rparents:
                for p in self.rparents:
                    p.color_update(propagate=True,ts=ts)

    def update_color(self,color):
        if self.color == color:
            self.color_count+=1
        else:
            self.color_count=0
            self.color = color

    @hybrid_property
    def since_last_update(self):
        now = datetime.now(timezone.utc).timestamp()
        difference = (now - (self.updated_on.timestamp() + self.get_timezone_offset()) )
        return difference 

# CONVERT=operators.custom_op('@')
# class UnitTest(Pumpsense):
#     __tablename__ = 'unit_test'
#     id = Column(Integer, primary_key=True)
#     name = Column(String)
#     value = Column(SiUnit)
#     # xx = column_property(value.op("@",'degF'))
#     converted = query_expression()

#     @hybrid_property
#     def convert(self):
#         print("convert")
#         return self.value
    
#     # @convert.inplace.getter
#     # @classmethod
#     # def _convert(cls):
#     #     print("convert getter")
#     #     return cls.value.op("@",'degF')
#     #     return cls.value.operate(CONVERT,cls.value,'degF')
    
#     @convert.expression
#     # @classmethod
#     def convert(cls):
#         print("_convert_expression")
#         return cls.value.op("@",'degF')

class GaugeSeeding(Pumpsense):
    __tablename__ = 'gauge_seedings'
    id = Column(Integer, primary_key=True)
    name = Column(String(256), nullable=False, unique=True)

class GaugeRange(Pumpsense):
    __tablename__ = 'gauge_ranges'
    id = Column(Integer, primary_key=True)
    name = Column(String(50))
    trend_meta_data_name = Column(String)
    trend_meta_data_id = Column(Integer,ForeignKey(TrendMetaData.id, ondelete='CASCADE'),nullable=True)
    pm_function_id = Column(Integer,ForeignKey(PmFunction.id, ondelete='CASCADE'),nullable=True)
    device_data_item_id = Column(Integer,ForeignKey(DeviceDataItem.id, ondelete='CASCADE'),nullable=True)
    color = Column(String(50))
    start = Column(Float,nullable=True)
    end = Column(Float,nullable=True)
    fps_start = Column(Float,nullable=True)
    fps_end = Column(Float,nullable=True)
    mks_start = Column(Float,nullable=True)
    mks_end = Column(Float,nullable=True)

    device_data_item = relationship(DeviceDataItem,uselist=False, foreign_keys=[device_data_item_id], back_populates="gauge_ranges")
    pm_function = relationship('PmFunction', uselist=False, foreign_keys=[pm_function_id], back_populates="gauge_ranges")
    trend_meta_data = relationship(TrendMetaData, uselist=False, foreign_keys=[trend_meta_data_id], back_populates="gauge_ranges")

class ForgotPassword(Pumpsense):
    __tablename__ = 'forgot_passwords'
    id = Column(Integer, primary_key=True)
    user_id = Column(Integer,ForeignKey(User.id, ondelete='CASCADE'),index=True)
    token = Column(String(64),unique=True)
    expires = Column(TIMESTAMP,nullable=False)
    used = Column(TIMESTAMP,nullable=True)

    user = relationship(User,uselist=False)

class ETLTracker(Pumpsense):
    __tablename__ = 'etl_trackers'
    id = Column(Integer, primary_key=True)
    etl_category = Column(String(50),unique=True)
    last_run = Column(TIMESTAMP, nullable=True)
    start_time = Column(TIMESTAMP, nullable=True)
    end_time = Column(TIMESTAMP, nullable=True)
    updated_on = Column(TIMESTAMP, nullable=False, default=datetime.utcnow, onupdate=datetime.utcnow)

    def set_last_run(self,start_time, end_time):
        self.last_run = datetime.now()
        self.start_time = start_time
        self.end_time = end_time


class MroIssue(Pumpsense):
    __tablename__ = 'mro_issues'
    id = Column(Integer, primary_key=True)
    pm_function_id = Column(Integer,ForeignKey(PmFunction.id, ondelete='CASCADE'),nullable=True, index=True)
    color = Column(String(20))
    description = Column(String(512))
    preventive_action = Column(String(512))


    pm_function = relationship('PmFunction', uselist=False, foreign_keys=[pm_function_id], back_populates="mro_issues")


class OtaFile(Pumpsense):
    __tablename__ = 'ota_files'
    id = Column(Integer, primary_key=True)
    product_line = Column(String(20))
    version = Column(String(20))
    updated_on = Column(TIMESTAMP, nullable=False, default=datetime.utcnow, onupdate=datetime.utcnow)
    file_name = Column(String(256), nullable=False, unique=True)
    file_path = Column(String(256), nullable=False, unique=True)
    file_system = Column(String(20))
    pi_version = Column(String(20))
    file_type = Column(String(20))