from ..module import bp
from pathlib import Path
import os
import json
import time
import gzip
import shutil
import shlex
import subprocess
from datetime import datetime
import logging
from mdhwebserver.appservices import database as db
from mdhwebserver import appconfigs

from http import HTTPStatus
from flask import (
    request,
    redirect,
    abort,
    send_file
)
from flask_jwt_extended import current_user
from mdhwebserver.wrappers.is_authorized import is_authorized

from mdhwebserver.builders.responses import JsonResponse
from mdhwebserver.builders.response_builder import SQLAlchemyResponseBuilder, MDHResponseBuilder
from mdhwebserver.builders import ValidationErrorResponseBuilder
from mdhwebserver.exceptions import ValidationException
from mdhwebserver.utils.listutils import dict_extract
from mdhwebserver import application

from sqlalchemy.exc import IntegrityError

logger = logging.getLogger('db')

# backup_directory = '/opt/novatec/data/backups'

@bp.route('/backup', methods=['POST'])
@is_authorized()
def backup():
    output = backup_db()
    return JsonResponse(MDHResponseBuilder(statusCode=HTTPStatus.OK.value, body=output))

@bp.route('/restore', methods=['POST'])
@is_authorized()
def restore():
    output = restore_db()
    return JsonResponse(MDHResponseBuilder(statusCode=HTTPStatus.OK.value))

@bp.route('/backups', methods=['GET'])
@is_authorized()
def get_backups():
    server_config = appconfigs.server
    backup_directory = server_config['backup_directory']
    if not Path(backup_directory).exists():
        output = ''
        return JsonResponse(MDHResponseBuilder(statusCode=HTTPStatus.OK.value, body=output))
    
    p = Path(backup_directory)
    files = list(p.glob('*'))
    output = []
    for file in files:
        print(file)
        file_name = file.parts[len(file.parts) - 1]
        file_size = sizeof_fmt(os.path.getsize(file))
        file_date = time.ctime(os.path.getmtime(file))
        output.append({'name': file_name, 'size': file_size, 'date': file_date})
    return JsonResponse(MDHResponseBuilder(statusCode=HTTPStatus.OK.value, body=output))

@bp.route('/backup/<file_name>', methods=['GET'])
@is_authorized()
def get_backup(file_name):
    server_config = appconfigs.server
    backup_directory = server_config['backup_directory']
    if Path(backup_directory+'/'+file_name).exists():
        # print(os.path.join(backup_directory, file_name))
        return send_file(os.path.join(backup_directory, file_name),'application/gzip')
    else:
        abort(HTTPStatus.NOT_FOUND)

@bp.route('/backup/<file_name>', methods=['DELETE'])
@is_authorized()
def delete_backup(file_name):
    server_config = appconfigs.server
    backup_directory = server_config['backup_directory']
    if Path(backup_directory+'/'+file_name).exists():
        os.remove(os.path.join(backup_directory, file_name))
        return JsonResponse(MDHResponseBuilder(statusCode=HTTPStatus.OK.value))
    else:
        abort(HTTPStatus.NOT_FOUND)

@bp.route('/backup/upload', methods=['POST'])
@is_authorized()
def upload_backup():
    if not request.files:
        abort(HTTPStatus.BAD_REQUEST)
    file = request.files['file']
    if file.filename == '':
        abort(HTTPStatus.BAD_REQUEST)
    if file and file.filename:
        server_config = appconfigs.server
        backup_directory = server_config['backup_directory']
        deleteOldBackups()
        filePath = os.path.join(backup_directory, file.filename)
        # create backup
        file.save(filePath)

        while not os.path.exists(filePath):
            time.sleep(1)
            
        if not is_gz_file(filePath):
            print('Not a gzip file')
            with open(filePath, 'rb') as f_in:
                with gzip.open(filePath+'.gz', 'wb') as f_out:
                    shutil.copyfileobj(f_in, f_out)
            
            if Path(filePath+'.gz').exists():
                os.remove(filePath)
        return JsonResponse(MDHResponseBuilder(statusCode=HTTPStatus.OK.value))
    else:
        abort(HTTPStatus.BAD_REQUEST)


def backup_db():
    deleteOldBackups()
    server_config = appconfigs.server
    backup_directory = server_config['backup_directory']
    if not Path(backup_directory).exists():
        Path(backup_directory).mkdir(parents=True)
    
    db_config = appconfigs.database
    filename = create_unique_filename(backup_directory, 'mdhdb', '.dump')
    print('filename', backup_directory+'/'+filename)
    pgdump_command = server_config['pgdump_command']
    f = open(backup_directory+'/'+filename.replace('.dump','.lock'), "x")
    backup_command = f"{pgdump_command} --username=postgres -e timescaledb -e unit --disable-triggers -Fc {db_config['db_name']} > {backup_directory}/{filename}"

    p = subprocess.Popen(backup_command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    p.communicate()
    with open(backup_directory+'/'+filename, 'rb') as f_in:
        with gzip.open(backup_directory+'/'+filename+'.gz', 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)
    
    if Path(backup_directory+'/'+filename+'.gz').exists():
        os.remove(backup_directory+'/'+filename)
    
    os.remove(backup_directory+'/'+filename.replace('.dump','.lock'))
    return ''

def restore_db():
    server_config = appconfigs.server
    backup_directory = server_config['backup_directory']
    psql_command = server_config['psql_command']
    pgrestore_command = server_config['pgrestore_command']
    if not Path(backup_directory).exists():
        abort(HTTPStatus.NOT_FOUND)
    files = list(Path(backup_directory).glob('*'))
    if len(files) == 0:
        abort(HTTPStatus.NOT_FOUND)
    if len(files) > 1:
        abort(HTTPStatus.BAD_REQUEST)
    file = files[0]
    file_name = file.parts[len(file.parts) - 1]
    db_config = appconfigs.database

    restore_file = open(backup_directory+'/db.restore', "x")

    orig_filename = file_name
    if is_gz_file(orig_filename):
        file_name = file_name.replace('.gz', '')
        unzipped_file = backup_directory+'/'+file_name
        
        with gzip.open(file, 'rb') as f_in:
            with open(unzipped_file, 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)
    
    # pg_restore isn't able to drop the database, so we need to do it manually
    #pgdrop_command = f"psql -h {db_config['host']} -U postgres -c 'DROP DATABASE {db_config['db_name']}'"
    pgdrop_command = f"{psql_command} -U postgres -c 'DROP DATABASE {db_config['db_name']} with (FORCE)'"
    p = subprocess.Popen(pgdrop_command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    p.communicate()
    
    # after dropping the database, we need to recreate it.
    pgdrop_command = f"{psql_command} -U postgres -c 'CREATE DATABASE {db_config['db_name']}'"
    p = subprocess.Popen(pgdrop_command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    p.communicate()

    # perform a series of steps for timescaledb
    # https://docs.timescale.com/self-hosted/latest/migration/entire-database/
    # added flag in pg_dump to include timescaledb extension, so this may no longer be needed.
    # pg_command = f"psql -U postgres -d {db_config['db_name']} -c 'CREATE EXTENSION IF NOT EXISTS timescaledb'"
    # p = subprocess.Popen(pg_command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    # p.communicate()

    # added flag in pg_dump to include timescaledb extension, so this may no longer be needed.
    # pg_command = f"psql -U postgres -d {db_config['db_name']} -c 'select timescaledb_pre_restore()'"
    # p = subprocess.Popen(pg_command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    # p.communicate()

    pg_restore_command = f"{pgrestore_command} --username=postgres --clean --create --disable-triggers -d {db_config['db_name']} {backup_directory}/{file_name}"
    p = subprocess.Popen(pg_restore_command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    p.communicate()

    # added flag in pg_dump to include timescaledb extension, so this may no longer be needed.
    # pg_command = f"psql -U postgres -d {db_config['db_name']} -c 'select timescaledb_post_restore()'"
    # p = subprocess.Popen(pg_command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    # p.communicate()

    if (is_gz_file(orig_filename)):
        os.remove(os.path.join(unzipped_file))

    os.remove(os.path.join(backup_directory+'/db.restore'))

    # trigger reboot of datahub in order to restart services after restore
    server_config = appconfigs.server
    reboot_command = server_config['reboot_command']
    reboot = f"{reboot_command}"
    p = subprocess.Popen(reboot, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    p.communicate()
    return ''

def sizeof_fmt(num, suffix="B"):
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1000.0:
            return f"{num:3.1f} {unit}{suffix}"
        num /= 1000.0
    return f"{num:.1f} Yi{suffix}"

# def allowed_file(filename):
#     return '.' in filename and \
#            filename.rsplit('.', 1)[1].lower() in appconfigs.ALLOWED_EXTENSIONS

def deleteOldBackups():
    # delete old backups
    server_config = appconfigs.server
    backup_directory = server_config['backup_directory']
    if Path(backup_directory).exists():
        files = list(Path(backup_directory).glob('*'))
        for file in files:
            os.remove(os.path.join(backup_directory, file.parts[len(file.parts) - 1]))

def is_gz_file(filepath):
    print('is_gz_file test')
    if filepath.endswith(".gz"):
        return True
    return False

def create_unique_filename(directory, filename, extension=".txt"):
    """
    Creates a unique filename with the current date and a sequence number if necessary.

    Parameters:
    - directory: The directory to check for existing files.
    - extension: The file extension to be appended. Defaults to ".txt".

    Returns:
    - A unique filename as a string.
    """
    # Get the current date in YYYYMMDD format
    date_str = datetime.now().strftime("%Y%m%d")
    # Initialize the sequence number
    seq = 0
    # Construct the initial filename
    new_filename = f"{filename}-{date_str}{extension}"
    
    # If the filename already exists, increment the sequence number
    # until a unique filename is found.
    while os.path.exists(os.path.join(directory, new_filename)):
        seq += 1
        new_filename = f"{filename}-{date_str}-{seq}{extension}"
    
    # Return the unique filename
    return new_filename