from flask import Flask, request, jsonify, send_file

import simplejson as json
from mdhwebserver.appservices import database as db
from mdhwebserver.utils.sqlalchemy import getModelByName
from mdhwebserver.builders.responses import JsonResponse
from mdhwebserver.builders.response_builder import SQLAlchemyResponseBuilder, MDHResponseBuilder, MDHErrorResponseBuilder
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.orm.collections import InstrumentedList

from mdhwebserver.models.models import (
    PmFunction,
    Assembly,
    SubassemblyInstance,
    TrendMetaData,
    GaugeRange,
)

from mdhwebserver.models.SiUnitModels import (
    SiUnitUnit,
    SiUnitPrefix,
)

from http import HTTPStatus

from ..gaugeeditor import bp

import time
from werkzeug.utils import secure_filename
import os
# import json
import shutil
import sqlalchemy
import tarfile

from .pmfunction import SetPmFunctionBasic, SetPmFunctionAdvanced, SetPmFunctionTrend, SetPmFunctionAddGaugeRanges, SetPmFunctionEditGaugeRanges
from .assembly import SetAssemblyBasic
from .subassembly import SetSubassemblyBasic

from sqlalchemy import func, inspect

from flask_jwt_extended import current_user
from mdhwebserver.wrappers.is_authorized import is_authorized

from mdhwebserver.modules.units.controllers.units import si_unit_convert

from mdhwebserver import appconfigs
from mergedeep import merge
from flask import abort

config = appconfigs.gaugeeditor

UPLOAD_FOLDER = config["upload_folder"]
ALLOWED_EXTENSIONS = {'svg'}

# Check for allowed file type/extension
def allowed_file(filename):
    return '.' in filename and filename.rsplit('.', 1)[1].lower() in ALLOWED_EXTENSIONS

# Replace filename with epoch time
def ParseFileName(image):
    # Get epoch time, for setting file name
    epoch_time = int(time.time())
    epoch_time_string = str(epoch_time)
    # Get file extension, for keeping file format intact
    extension = image.filename.split(".")
    extension = extension[len(extension) - 1]
    # Set new filename
    filename = epoch_time_string + "." + extension

    return filename

@bp.put('/assembly/<id>')
@is_authorized()
def update_assembly(id=None):
    try:
        # Get form data
        name = request.form.get('name')
        chosen_image = request.form.get('chosen_image')
        buildable = request.form.get('buildable')
        filename = None

        if chosen_image:
            chosen_image = os.path.basename(chosen_image)
            chosen_image, ext = os.path.splitext(chosen_image)

        # Check for image in request
        if 'image' in request.files:
            image = request.files['image']
            filename = ParseFileName(image)

            # Check image extension, then save file
            if image and allowed_file(filename):
                filename = secure_filename(filename)
                image.save(os.path.join(UPLOAD_FOLDER, filename))
            else:
                return JsonResponse({"success": False, "error": "File type not allowed"}, status=HTTPStatus.BAD_REQUEST)
    except Exception as e:
        print("Error in first try block")
        print(e)
        return JsonResponse({"success": False, "error": str(e)}, status=HTTPStatus.INTERNAL_SERVER_ERROR)

    try:
        # For Assembly DB model
        assembly = Assembly.where(id=id).first()
        # Settings under Basic tab in frontend
        SetAssemblyBasic(assembly, name, request, chosen_image, filename, buildable)
        # Commit the changes to the database
        db.session.commit()
    except Exception as e:
        print(e)
        return JsonResponse({"success":False, "error": str(e)}, status=HTTPStatus.INTERNAL_SERVER_ERROR)

    if 'image' in request.files:
        return JsonResponse({"success":True, "image_file": "web-images/" + filename})
    else:
        return JsonResponse({"success":True})

@bp.put('/subassembly/<id>')
@is_authorized()
def update_subassembly(id=None):
    try:
        # Get form data
        name = request.form.get('name')
    except Exception as e:
        print("Error in first try block")
        print(e)
        return JsonResponse({"success": False, "error": str(e)}, status=HTTPStatus.INTERNAL_SERVER_ERROR)

    try:
        # For SubassemblyInstance DB model
        subassembly = SubassemblyInstance.where(id=id).first()
        # Settings under Basic tab in frontend
        SetSubassemblyBasic(subassembly, name)
        subassembly.save()
        # Commit the changes to the database
        db.session.commit()
    except Exception as e:
        print(e)
        return JsonResponse({"success":False, "error": str(e)}, status=HTTPStatus.INTERNAL_SERVER_ERROR)

    return JsonResponse({"success":True})

@bp.put('/pmfunction/<id>')
@is_authorized()
def update_pmfunction(id=None):
    try:
        # Get form data
        name = request.form.get('name')
        chosen_image = request.form.get('chosen_image')
        filename = None
        new_gauge_data = request.form.get('new_gauge_data')
        new_gauge_data = json.loads(new_gauge_data)
        add_gauge_range_data = request.form.get('add_gauge_range_data')
        add_gauge_range_data = json.loads(add_gauge_range_data)
        edit_gauge_range_data = request.form.get('edit_gauge_range_data')
        edit_gauge_range_data = json.loads(edit_gauge_range_data)
        print("New Gauge Data", new_gauge_data)

        if chosen_image != "":
            chosen_image = os.path.basename(chosen_image)
            chosen_image, ext = os.path.splitext(chosen_image)

        # Check for image in request
        if 'image' in request.files:
            image = request.files['image']
            filename = ParseFileName(image)

            # Check image extension, then save file
            if image and allowed_file(filename):
                filename = secure_filename(filename)
                print(UPLOAD_FOLDER)
                print(filename)
                image.save(os.path.join(UPLOAD_FOLDER, filename))
            else:
                return JsonResponse({"success": False, "error": "File type not allowed"}, status=HTTPStatus.BAD_REQUEST)
    except Exception as e:
        print("Error in first try block")
        print(e)
        return JsonResponse({"success": False, "error": str(e)}, status=HTTPStatus.INTERNAL_SERVER_ERROR)

    try:
        # For PmFunction DB model
        pmfunction = PmFunction.where(id=id).first()
        # Settings under Advanced tab in frontend
        SetPmFunctionAdvanced(pmfunction, new_gauge_data)
        # Settings under Basic tab in frontend
        SetPmFunctionBasic(pmfunction, name, request, chosen_image, filename)
        # Settings under Trend tab in frontend
        SetPmFunctionTrend(new_gauge_data)
        # Settings under Ranges tab in frontend
        add_gauge_ranges = SetPmFunctionAddGaugeRanges(pmfunction, add_gauge_range_data)
        SetPmFunctionEditGaugeRanges(pmfunction, edit_gauge_range_data)
        # Save PmFunction model
        pmfunction.save()
        # Commit the changes to the database
        db.session.commit()
    except Exception as e:
        print(e)
        return JsonResponse({"success":False, "error": str(e)}, status=HTTPStatus.INTERNAL_SERVER_ERROR)

    if 'image' in request.files:
        return JsonResponse({"success":True, "image_file": "web-images/" + filename, "add_gauge_ranges": add_gauge_ranges})
    else:
        return JsonResponse({"success":True, "add_gauge_ranges": add_gauge_ranges})

# @bp.post('/update')
# @is_authorized()
# def update():
#     try:
#         # Get form data
#         model = request.form.get('model')
#         id = request.form.get('id')
#         name = request.form.get('name')
#         chosen_image = request.form.get('chosen_image')
#         filename = None
#         if model == "PmFunction":
#             new_gauge_data = request.form.get('new_gauge_data')
#             new_gauge_data = json.loads(new_gauge_data)
#             print("New Gauge Data", new_gauge_data)

#         if chosen_image:
#             chosen_image = os.path.basename(chosen_image)
#             chosen_image, ext = os.path.splitext(chosen_image)

#         # Check for image in request
#         if 'image' in request.files:
#             image = request.files['image']
#             filename = ParseFileName(image)

#             # Check image extension, then save file
#             if image and allowed_file(filename):
#                 filename = secure_filename(filename)
#                 image.save(os.path.join(UPLOAD_FOLDER, filename))
#             else:
#                 return JsonResponse({"success": False, "error": "File type not allowed"}, status=HTTPStatus.BAD_REQUEST)

#     except Exception as e:
#         print("Error in first try block")
#         print(e)
#         return JsonResponse({"success": False, "error": str(e)}, status=HTTPStatus.INTERNAL_SERVER_ERROR)

#     try:
#         # For Assembly DB model
#         if model == "Assembly":
#             assembly = Assembly.where(id=id).first()
#             # Settings under Basic tab in frontend
#             SetAssemblyBasic(assembly, name, request, chosen_image, filename)
#         # For SubassemblyInstance DB model
#         if model == "SubassemblyInstance":
#             subassembly = SubassemblyInstance.where(id=id).first()
#             # Settings under Basic tab in frontend
#             SetSubassemblyBasic(subassembly, name)
#             subassembly.save()
#         # For PmFunction DB model
#         if model == "PmFunction":
#             pmfunction = PmFunction.where(id=id).first()
#             # Settings under Advanced tab in frontend
#             SetPmFunctionAdvanced(pmfunction, new_gauge_data)
#             # Settings under Basic tab in frontend
#             SetPmFunctionBasic(pmfunction, name, request, chosen_image, filename)
#             # Settings under Trend tab in frontend
#             SetPmFunctionTrend(new_gauge_data)
#             # Save PmFunction model
#             pmfunction.save()
#         # Commit the changes to the database
#         db.session.commit()
#     except Exception as e:
#         print(e)
#         return JsonResponse({"success":False, "error": str(e)}, status=HTTPStatus.INTERNAL_SERVER_ERROR)

#     if 'image' in request.files:
#         return JsonResponse({"success":True, "image_file": "web-images/" + filename})
#     else:
#         return JsonResponse({"success":True})


@bp.get('/list_files')
@is_authorized()
def list_files():
    # Specify the directory path you want to list files from
    directory_path = config["list_files_path"]

    # Get the list of files in the directory
    files = os.listdir(directory_path)

    # Create a JSON response
    response_data = {'files': files}

    # Convert the dictionary to JSON and return the response
    return JsonResponse(response_data)

@bp.route('/<pm_id>', methods=['delete'])
@is_authorized()
def delete_gauge(pm_id=None):
    PmFunction.query.filter(PmFunction.id==pm_id).delete()
    db.session.commit()
    return JsonResponse({'success':True})

@bp.route('/trend/<trend_id>', methods=['delete'])
@is_authorized()
def delete_trend(trend_id=None):
    TrendMetaData.query.filter(TrendMetaData.id==trend_id).delete()
    db.session.commit()
    return JsonResponse({'success':True})

from sqlalchemy.orm import selectinload, joinedload

def fetch_seeding_data(assembly_id):
    assembly = Assembly.where(id=assembly_id).options(
            selectinload(Assembly.subassemblies).options(
                selectinload(SubassemblyInstance.pm_functions).options(
                    selectinload(PmFunction.gauge_group),
                    selectinload(PmFunction.trend_meta_data),
                    selectinload(PmFunction.gauge_ranges),
                    selectinload(PmFunction.etl_gauge),
                    selectinload(PmFunction.mro_issues)
                )
            )
        ).first()
    return assembly

def build_seeding_files_from_assembly(assembly_data, seed_root_path):
    seed_folder = seed_root_path   # This is so we can make a unique directory later and avoid race conditions.
    image_folder = os.path.join(seed_folder,'images')
    os.makedirs(seed_folder, exist_ok=True)
    os.makedirs(image_folder, exist_ok=True)

    try:
        copy_image_to_zip(assembly_data.image,image_folder)
    except Exception as e:
        print(f"Error copying image: {e}")

    for subassembly in assembly_data.subassemblies:
        for pm_function in subassembly.pm_functions:
            try:
                copy_image_to_zip(pm_function.gauge_base_image,image_folder)
            except Exception as e:
                print(f"Error copying image: {e}")

    with open(seed_folder+'/data.json', 'w') as file:
        json.dump({"data": assembly_data._to_dict()}, file, indent=4, default=str)
    return seed_folder

@bp.route('/fetch_seeding/<id>', methods=['get'])
@is_authorized()
def fetch_seeding(id=None):
    assembly_data = fetch_seeding_data(id)
    if assembly_data is None:
        abort(404,"Assembly not found")
    try:
        seed_folder = build_seeding_files_from_assembly(assembly_data,config['fetch_seeding_path'])
    except Exception as e:
        return JsonResponse(MDHErrorResponseBuilder(str(e)))

    try:
        output_file = config['base_path']+"/fetch_seeding.tar.gz"
        make_tar_gz(output_file, seed_folder)
    except Exception as e:
        shutil.rmtree(seed_folder)
        return JsonResponse(MDHErrorResponseBuilder(str(e)))

    try:
        shutil.rmtree(seed_folder)
    except Exception as e:
        return JsonResponse(MDHErrorResponseBuilder(str(e)))

    return send_file(output_file, as_attachment=True)

def copy_image_to_zip(image,destination_directory_path):
    if "web-images" in image:
        filename = image.split("/")
        filename = filename[-1]
        source_file_path = config['upload_folder'] + filename                  
        destination_file_path = destination_directory_path + filename
        shutil.copy(source_file_path, destination_file_path)

def make_tar_gz(output_filename, source_dir):
    with tarfile.open(output_filename, "w:gz") as tar:
        tar.add(source_dir, arcname=os.path.basename(source_dir))

@bp.put('/gaugerange/<gr_id>')
@is_authorized()
def edit_gaugerange(gr_id=None):
    # Get form data
    name = request.form.get('name')
    trend_data_id = request.form.get('trend_id')
    trend_data_name = request.form.get('trend_name')
    device_data_id = request.form.get('device_data_id')
    color = request.form.get('color')
    range_start = float(request.form.get('range_start'))
    range_end = float(request.form.get('range_end'))

    GaugeRange = GaugeRange.where(id=gr_id).first()
    if GaugeRange is None:
        return jsonify({"error": "GaugeRange not found"}), 404

    PmFunc = PmFunction.where(id=GaugeRange.pm_function_id).first()
    if PmFunc is None:
        return jsonify({"error": "PmFunction not found"}), 404

    convert_units = { }
    mks_start = fps_start = convert_units['start'] = range_start
    mks_end = fps_end = convert_units['end'] = range_end

    if PmFunc.trend_meta_mapped[trend_data_name].fps_units != PmFunc.trend_meta_mapped[trend_data_name].mks_units:
        if PmFunc.trend_meta_mapped[trend_data_name].units != PmFunc.trend_meta_mapped[trend_data_name].fps_units:
            converted = si_unit_convert(PmFunc.trend_meta_mapped[trend_data_name].units,PmFunc.trend_meta_mapped[trend_data_name].fps_units,convert_units)
            fps_start = converted.get("start")
            fps_end = converted.get("end")
        else:
            converted = si_unit_convert(PmFunc.trend_meta_mapped[trend_data_name].units,PmFunc.trend_meta_mapped[trend_data_name].mks_units,convert_units)
            mks_start = converted.get("start")
            mks_end = converted.get("end)")

    try:
        GaugeRange.name = name
        GaugeRange.trend_meta_data_name = trend_data_name
        GaugeRange.trend_meta_data = PmFunc.trend_meta_mapped[trend_data_name]
        GaugeRange.pm_function = PmFunc
        GaugeRange.color = color
        GaugeRange.start = range_start
        GaugeRange.end = range_end
        GaugeRange.fps_start = fps_start
        GaugeRange.fps_end = fps_end
        GaugeRange.mks_start = mks_start
        GaugeRange.mks_end = mks_end

        GaugeRange.save()
        db.session.commit()
    except Exception as e:
        print(e)
        return jsonify({"error": e}), 500
        
    return jsonify({"success": True}), 200

@bp.delete('/gaugerange/<gr_id>')
@is_authorized()
def delete_gaugerange(gr_id=None):
    try:
        GaugeRange.query.filter(GaugeRange.id==gr_id).delete()
        db.session.commit()
    except Exception as e:
        return  jsonify({"error": e}), 500

    return jsonify({"success": True}), 200


# @bp.get('/list_units')
# def list_units():
#     units=SiUnitUnit().all()
#     rb = SQLAlchemyResponseBuilder(units)
#     return JsonResponse(rb,content_type='application/json')

# @bp.get('/list_prefixes')
# def list_prefixes():
#     prefixes=SiUnitPrefix().all()
#     rb = SQLAlchemyResponseBuilder(prefixes)
#     return JsonResponse(rb,content_type='application/json')

# @bp.post('/unit_lookup')
# def unit_lookup():
#     data = request.json
#     search = data.get('search')

#     prefixes = SiUnitPrefix.query.filter(SiUnitPrefix.prefix.ilike(search)).all()
#     units = SiUnitUnit.query.filter(SiUnitUnit.name.ilike(search)).all()
#     data = []

#     for prefix in prefixes:
#         data.append(prefix.prefix)
#     for unit in units:
#         data.append(unit.name)

#     # rb = SQLAlchemyResponseBuilder(data)
#     # return JsonResponse(rb,content_type='application/json')

#     return jsonify({"success": True, "data": data})

# @bp.post('/unit_lookup')
# def unit_lookup():
#     data = request.json
#     search = data.get('search')

#     # Performing a cross join and filtering based on concatenated values
#     results = db.session.query(SiUnitPrefix, SiUnitUnit).filter(
#         func.concat(SiUnitPrefix.prefix, SiUnitUnit.name).ilike(f"%{search}%")
#     ).all()

#     data = []
#     for prefix, unit in results:
#         combined_name = prefix.prefix + unit.name
#         data.append(combined_name)

#     if len(data) <= 100:
#         return jsonify({"success": True, "data": {"show": True, "data": data}})

#     return jsonify({"success": True, "data": {"show": False, "data": []}})