import datetime
from ..module import bp
from mdhwebserver.models.models import OtaFile, Datahub
from os import getenv as env
import os
import json
import time
import re

import logging

from http import HTTPStatus
from flask import (
    request,
    redirect,
    abort,
    send_file
)
from uuid import uuid4
import urllib.parse

from datetime import datetime, timedelta, timezone
from mdhwebserver.wrappers.is_authorized import is_authorized
from mdhwebserver.appservices import database as db
from mdhwebserver.builders.responses import JsonResponse
from mdhwebserver.builders.response_builder import SQLAlchemyResponseBuilder, MDHResponseBuilder
from mdhwebserver.builders import ValidationErrorResponseBuilder
from mdhwebserver.exceptions import ValidationException
from mdhwebserver.utils.listutils import dict_extract
from mdhwebserver import application
from mdhwebserver import appconfigs
from sqlalchemy.orm import joinedload
from sqlalchemy.exc import IntegrityError

logger = logging.getLogger('ota')


@bp.route('/sysupdate/<mac_address>', methods=['GET'])
def download_ota_file(mac_address):
    datahub = Datahub.query.filter(Datahub.mac_address == mac_address).options(joinedload(Datahub.ota_file)).first()
    # print(datahub.ota_downloaded_on)
    # print(datetime.now() - timedelta(minutes=5))
    if datahub.ota_file and (datahub.ota_downloaded_on is None or datahub.ota_downloaded_on < datetime.now() - timedelta(minutes=5)):
        datahub.ota_downloaded_on = datetime.now()
        datahub.save()
        db.session.commit()
        return send_file(datahub.ota_file.file_path,'application/gzip')
    else:
        abort(HTTPStatus.NOT_FOUND)
    

@bp.route('/upload', methods=['POST'])
@is_authorized()
def upload_ota_file():
    if not request.files:
        abort(HTTPStatus.BAD_REQUEST)
    file = request.files['file']
    if file.filename == '':
        abort(HTTPStatus.BAD_REQUEST)
    if file and file.filename:
        server_config = appconfigs.server
        user_files_upload_directory = server_config['user_files_upload_directory']
        upload_directory = user_files_upload_directory+'/ota/'
        if not os.path.exists(upload_directory):
            os.makedirs(upload_directory)
        file_path = os.path.join(upload_directory, file.filename)
        print(file_path)
        file.save(file_path)

        ota_file = OtaFile().where(file_name=file.filename).first()
        if not ota_file:        
            ota_file = OtaFile().create(**{
                'created_on': datetime.now(),
                'file_name': file.filename,
                'file_path': file_path,
            })

        ota_file.updated_on = datetime.now()

        product_line, version, pi_version, file_system, file_type = extract_info(file.filename)
        if product_line and version and pi_version and file_system and file_type:
            ota_file.product_line = product_line
            ota_file.version = version
            ota_file.pi_version = pi_version
            ota_file.file_system = file_system
            ota_file.file_type = file_type


        print(product_line, version, pi_version, file_system)
        db.session.commit()

        return JsonResponse(MDHResponseBuilder(statusCode=HTTPStatus.OK.value))
    else:
        abort(HTTPStatus.BAD_REQUEST)

def extract_info(filename):
    pattern = r'^[a-zA-Z]+-(\d+)\.(\d+)-[a-zA-Z0-9-]+-bcm27xx-bcm271\d-(rpi-\d+)-([^-]+)-([^.]+)\.img\.gz'
    match = re.match(pattern, filename)
    
    if match:
        product_line = match.group(1)
        version = match.group(2)
        pi_version = match.group(3)
        file_system = match.group(4)
        file_type = match.group(5)
        return product_line, version, pi_version, file_system, file_type
    else:
        return None, None, None, None, None

# Example usage
# filename1 = "mdh-1201.202406191-0-bcm27xx-bcm2711-rpi-4-ext4-sysupgrade.img.gz"
# filename2 = "plc-1302.202306121-1-bcm27xx-bcm2711-rpi-3-squashfs-factory.img.gz"
# filename3 = "expac-1620.202419061-r22193-fd61006067-bcm27xx-bcm2710-rpi-3-ext4-sysupgrade.img.gz"
# product_line1, version1, pi_version1, file_system1 = extract_info(filename1)
# product_line2, version2, pi_version2, file_system2 = extract_info(filename2)
# print("Filename 1 - Product Line:", product_line1, "Version:", version1, "Pi Version:", pi_version1, "File System:", file_system1)
# print("Filename 2 - Product Line:", product_line2, "Version:", version2, "Pi Version:", pi_version2, "File System:", file_system2)



@bp.route('/files', methods=['GET'])
@is_authorized()
def get_ota_files():
    ota_files = OtaFile.query.order_by(OtaFile.created_on.desc()).all()
    return JsonResponse(SQLAlchemyResponseBuilder(ota_files).camel(False))

@bp.route('/file/<id>', methods=['DELETE'])
@is_authorized()
def delete_ota_file(id):
    ota_file = OtaFile.query.filter(OtaFile.id == id).first()
    if ota_file:
        filePath = ota_file.file_path
        os.remove(filePath)
        db.session.delete(ota_file)
        db.session.commit()
        return JsonResponse(MDHResponseBuilder(statusCode=HTTPStatus.OK.value))
    else:
        return JsonResponse(MDHResponseBuilder(statusCode=HTTPStatus.NOT_FOUND.value))
    

@bp.route('/assign', methods=['POST'])
@is_authorized()
def assign_ota_file():
    data = request.get_json()
    if not data:
        abort(HTTPStatus.BAD_REQUEST)
    ota_file_id = data.get('ota_file_id')
    datahubs = data.get('datahubs')
    if not datahubs or not ota_file_id:
        abort(HTTPStatus.BAD_REQUEST)
    ota_file = OtaFile.query.filter(OtaFile.id == ota_file_id).first()
    if not ota_file:
        abort(HTTPStatus.NOT_FOUND)

    datahubs = [int(float(c)) for c in datahubs]
    if datahubs is not None and len(datahubs) > 0:
        # print("add datahubs: ", datahubs)
        ota_datahubs = Datahub.filter(Datahub.id.in_(datahubs)).all()
        ota_file.datahubs.extend(ota_datahubs)
        for datahub in ota_datahubs:
            datahub.assigned_version = ota_file.product_line + '.' + ota_file.version
        db.session.commit()
        return JsonResponse(MDHResponseBuilder(statusCode=HTTPStatus.OK.value))
    else:
        abort(HTTPStatus.BAD_REQUEST)

