#####################################
# Pumpsense API routes for devices  #
#####################################

from ..pumpsense import bp

import logging
import json
from json.decoder import JSONDecodeError

import jsonpickle
from http import HTTPStatus
from flask import (
    request,
    abort
)
from flask_jwt_extended import current_user
from mdhwebserver.wrappers.is_authorized import is_authorized

from mdhwebserver.appservices import database as db
from mdhwebserver import appconfigs
from sqlalchemy.orm.attributes import flag_modified
import sqlalchemy as sa
from mdhwebserver.builders.response_builder import SQLAlchemyResponseBuilder 
from mdhwebserver.builders.response_builder import MDHResponseBuilder
from mdhwebserver.utils.listutils import (
    dict_map_extract, 
    dict_extract
)
from mdhwebserver.exceptions.pumpsense_error import (
    PumpsenseError,
    MalformedPayloadError
)
from mdhwebserver.models.models import (
    Gauge,
    Datahub,
    Assembly
)

from mdhwebserver.models.models import User

from ..tasks.tasks import submit_task

from mdhwebserver.builders.responses import JsonResponse


logger = logging.getLogger('devices')



@bp.route('/gauge', methods=['POST'])
def ingest_gauge_data():
    try:
        submit_task.delay(request.data.decode('UTF-8'))
        response = JsonResponse(MDHResponseBuilder(statusCode=HTTPStatus.OK.value))

    except:
        logger.error("exception occurred", exc_info=True)
        # response = get_default_err_resp_obj()
        abort(500,"Exception Occured")
    
    return response



@bp.route('/gauges', methods=['GET'])
def get_gauges():
    gauges = Gauge.with_joined(Gauge.device_data_items).all()

    rb = SQLAlchemyResponseBuilder(gauges)
    return JsonResponse(rb)


#We need to make a generic objectable delete, but this will work for now.
@bp.route('/gauge/<machine_id>', methods=['DELETE'])
def delete_gauge(machine_id=None):
    G = Gauge.query.filter(Gauge.machine_id==machine_id).delete()
    db.session.commit()
    return JsonResponse({'success':True})

@bp.route('/slave', methods=['PUT'])
def register_slave_device():
    try:
        try:
            payload = json.loads(request.data)
            mac = payload['macAddress']
            values = dict_map_extract(payload,{
                'macAddress':'mac_address','ipAddress':'ip_address','onboarded':'onboarded','currentVersion':'current_version','assignedVersion':'assigned_version', 
		'hubId':'hub_id','baseUrl':'base_url','name':'name','status':'status',
                'cloudId':'cloud_id',
                'serialNo':'serial_no'
            })

            print("\n\n\nValues are: ",values)
            datahub = Datahub.updateOrCreate({'mac_address':mac},**values)
            if not datahub.name:	# This is kludgie but we want to set a default name if the device doesn't have one.
                print("No datahub name")
                datahub.name = payload.get('name', f"datahub_{mac.replace(':', '')}") 
            machineList = payload.get('machineList')
            if machineList is not None:
                  for mvalues in machineList:
                       mid = mvalues['machineId']
                       machine_name=mvalues.get('machine_name','')
                       package_serial_no = mvalues.get('package_serial_no',None)
                       if mid in datahub.machines_mapped:
                             datahub.machines_mapped[mid].fill(**{'onboarded':mvalues['onboarded'],'package_serial_no':package_serial_no})
                       else:
                             G=Gauge.query.filter(Gauge.machine_id==mid).first()
                             if G:
                                 G.fill(**{'onboarded':mvalues['onboarded'],'datahub_id':datahub.id,'package_serial_no':package_serial_no})
                             else:
                                 assembly = Assembly.query.filter(Assembly.name=='unknown').first() 
                                 datahub.machines_mapped[mid] = Gauge(**{
                                    'machine_id':mvalues['machineId'],
                                    'onboarded':mvalues['onboarded'],
                                    'assembly_id':assembly.id,
                                    'machine_name': machine_name,
                                    'package_serial_no':package_serial_no
                                 })
                       #datahub.machines.append(Machine(**{'machine_id':mvalues['machineId'],'onboarded':mvalues['onboarded']}))

            if (datahub.assigned_version is not None and datahub.assigned_version == datahub.current_version):
                datahub.assigned_version = None
                datahub.ota_file_id = None
            flag_modified(datahub,'ip_address')	# Forces it to think it has an updated field so that onupdated_on gets updated.

            db.session.commit()
            return JsonResponse({"success":True})
        except JSONDecodeError:
            raise MalformedPayloadError(
                        request.data, 
                        'Payload is not in the expected format')
    
    except PumpsenseError as e:
        logger.error(e.message, exc_info=True)
        abort(HTTPStatus.BAD_REQUEST.value,e.message)

    except:
        logger.error("exception occurred", exc_info=True)
        abort(500,"Exception Occured")


@bp.route('/slave/<mac_addr>', methods=['DELETE'])
@is_authorized()
def delete_slave_device(mac_addr):
    G = Datahub.query.filter(Datahub.mac_address==mac_addr).delete()
    db.session.commit()
    return JsonResponse({'success':True})

@bp.route('/slave/<mac_addr>', methods=['GET'])
@is_authorized()
def get_slave_device(mac_addr):
    dh = Datahub.query.filter(Datahub.mac_address==mac_addr).first();
    rb = SQLAlchemyResponseBuilder(dh)
    return JsonResponse(rb)

@bp.route('/slaves', methods=['GET'])
@is_authorized()
def get_all_slave_devices():
    try:
        onboarded = request.args.get('onboarded', type=str)
        if onboarded is not None:
            onboarded = onboarded.lower()
            is_onboarded = (onboarded == 'true' or onboarded == '1')
            dh = Datahub.query.filter_by(onboarded=is_onboarded).all();
        else:
            dh = Datahub.all();
        
        rb = SQLAlchemyResponseBuilder(dh).camel(False)
        return JsonResponse(rb)
    
    except PumpsenseError as e:
        logger.error(e.message, exc_info=True)
        abort(HTTPStatus.BAD_REQUEST.value,e.message)

    except:
        logger.error("exception occurred", exc_info=True)
        abort(500,"Exception Occured")
        
@bp.route('/gauge', methods=['PUT'])
@is_authorized()
def update_gauge():
    try:
            payload = json.loads(request.data)
            values = dict_extract(payload,{'machine_id'})
            values['machine_name'] = payload.get('machine_name')
            values['onboarded'] = payload.get('onboarded')
            if 'pump_number' in payload.keys():
                values['pump_number'] = payload.get('pump_number')
            id = values['machine_id']
            print("machine id is: ",id)

            gauge = Gauge().where(machine_id=id).options(sa.orm.joinedload(Gauge.heatmap_item)).first()
            # print(values)
            gauge.fill(**values)
            if gauge.heatmap_item:
                print('Updating heatmap item name: ', gauge.heatmap_item.name)
                gauge.heatmap_item.name = values['machine_name']
                print('Updated heatmap item name: ', gauge.heatmap_item.name)
            db.session.commit()
    except:
        logger.error("exception occurred", exc_info=True)
        abort(400,"Error modifying gauge")
    return JsonResponse(MDHResponseBuilder(statusCode=HTTPStatus.OK.value))
